/*
 * Direct3D8 textures
 * Copyright (c) 2002, TransGaming Technologies Inc.
 */

#include "config.h"

#include "winbase.h"
#include "winuser.h"
#include "d3d8.h"
#include "debugtools.h"

#include "d3d8_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define TEXTURE_HANDLE(This) ((DWORD)&((IDirect3DSurface8Impl*)This->surf[0])->t)

HRESULT WINAPI Direct3DTexture8_QueryInterface(LPDIRECT3DTEXTURE8 iface,
					       REFIID riid,
					       LPVOID* obj)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->(%s,%p)\n", iface, debugstr_guid(riid), obj);
  if (IsEqualGUID(&IID_IUnknown, riid) ||
      IsEqualGUID(&IID_IDirect3DResource8, riid) ||
      IsEqualGUID(&IID_IDirect3DBaseTexture8, riid) ||
      IsEqualGUID(&IID_IDirect3DTexture8, riid)) {
    This->ref++;
    *obj = This;
    return S_OK;
  }
  else {
    FIXME("(%p)->(%s): no interface", iface, debugstr_guid(riid));
    return E_NOINTERFACE;
  }
}

ULONG WINAPI Direct3DTexture8_AddRef(LPDIRECT3DTEXTURE8 iface)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->()\n", iface);
  return ++This->ref;
}

ULONG WINAPI Direct3DTexture8_Release(LPDIRECT3DTEXTURE8 iface)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  unsigned u;

  TRACE("(%p)->()\n", iface);
  if (--This->ref) return This->ref;

  for (u=0; u<This->Levels; u++) {
    IDirect3DSurface8Impl* surf = (IDirect3DSurface8Impl*)This->surf[u];
    if (This->parent->demolish_texture)
      This->parent->demolish_texture(This->parent->ip_ctx,
				     (DWORD)&surf->t);
    IDirect3DSurface8_Release(This->surf[u]);
  }

  /* take the texture out of the linked list */
  if (This->prev_device)
    This->prev_device->next_device = This->next_device;
  else /* if there is no previous, this must be head */
    This->parent->textures = This->next_device;

  if (This->next_device)
    This->next_device->prev_device = This->prev_device;
  
  HeapFree(GetProcessHeap(), 0, This);
  return 0;
}

HRESULT WINAPI Direct3DTexture8_GetDevice(LPDIRECT3DTEXTURE8 iface,
					  IDirect3DDevice8** ppDevice)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  IDirect3DDevice8* pDevice;

  TRACE("(%p)->(%p)\n", iface, ppDevice);
  pDevice = (IDirect3DDevice8*)This->parent;
  IDirect3DDevice8_AddRef(pDevice);
  *ppDevice = pDevice;

  TRACE("=> %p\n", *ppDevice);
  return D3D_OK;
}

VOID WINAPI Direct3DTexture8_PreLoad(LPDIRECT3DTEXTURE8 iface)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXBLT, 0, {1} };
  D3DHAL_DP2TEXBLT data;

  TRACE("(%p)->()\n", iface);
  Direct3DTexture8_UpdateDirty(iface, FALSE);
  data.dwDDDestSurface = 0;
  data.dwDDSrcSurface = TEXTURE_HANDLE(This);
  data.pDest.x = 0;
  data.pDest.y = 0;
  data.rSrc.left   = 0;
  data.rSrc.top    = 0;
  data.rSrc.right  = This->Width;
  data.rSrc.bottom = This->Height;
  data.dwFlags = 0;
  This->parent->emit_dp2(This->parent->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

D3DRESOURCETYPE WINAPI Direct3DTexture8_GetType(LPDIRECT3DTEXTURE8 iface)
{
  TRACE("(%p)->()\n", iface);
  return D3DRTYPE_TEXTURE;
}

DWORD WINAPI Direct3DTexture8_GetLevelCount(LPDIRECT3DTEXTURE8 iface)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->()\n", iface);
  return This->Levels;
}

HRESULT WINAPI Direct3DTexture8_GetLevelDesc(LPDIRECT3DTEXTURE8 iface,
					     UINT Level,
					     D3DSURFACE_DESC* pDesc)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->(%d,%p)\n", iface, Level, pDesc);
  return IDirect3DSurface8_GetDesc(This->surf[Level], pDesc);
}

HRESULT WINAPI Direct3DTexture8_GetSurfaceLevel(LPDIRECT3DTEXTURE8 iface,
						UINT Level,
						IDirect3DSurface8** ppSurfaceLevel)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->(%d,%p)\n", iface, Level, ppSurfaceLevel);
  *ppSurfaceLevel = This->surf[Level];
  IDirect3DSurface8_AddRef(*ppSurfaceLevel);
  return D3D_OK;
}

HRESULT WINAPI Direct3DTexture8_LockRect(LPDIRECT3DTEXTURE8 iface,
					 UINT Level,
					 D3DLOCKED_RECT* pLockedRect,
					 CONST RECT* pRect,
					 DWORD Flags)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  HRESULT hr;
  TRACE("(%p)->(%d,%p,%p,%lx)\n", iface, Level,
	pLockedRect, pRect, Flags);
  hr = IDirect3DSurface8_LockRect(This->surf[Level],
				  pLockedRect,
				  pRect,
				  Flags);
  return hr;
}

HRESULT WINAPI Direct3DTexture8_UnlockRect(LPDIRECT3DTEXTURE8 iface,
					   UINT Level)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->(%d)\n", iface, Level);
  return IDirect3DSurface8_UnlockRect(This->surf[Level]);
}

HRESULT WINAPI Direct3DTexture8_AddDirtyRect(LPDIRECT3DTEXTURE8 iface,
					     CONST RECT* pDirtyRect)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  TRACE("(%p)->(%p)\n", iface, pDirtyRect);
  if (pDirtyRect)
    UnionRect(&This->DirtyRect, &This->DirtyRect, pDirtyRect);
  else
    SetRect(&This->DirtyRect, 0, 0, This->Width, This->Height);
  This->parent->dirty_textures = TRUE;
  return D3D_OK;
}

static ICOM_VTABLE(IDirect3DTexture8) D3DT8_Vtbl =
{
  ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
  Direct3DTexture8_QueryInterface,
  Direct3DTexture8_AddRef,
  Direct3DTexture8_Release,
  Direct3DTexture8_GetDevice,
  0xdead4005,
  0xdead4006,
  0xdead4007,
  0xdead4008, /* SetPriority */
  0xdead400a,
  Direct3DTexture8_PreLoad,
  Direct3DTexture8_GetType,
  0xdead400d, /* SetLOD */
  0xdead400e,
  Direct3DTexture8_GetLevelCount,
  Direct3DTexture8_GetLevelDesc,
  Direct3DTexture8_GetSurfaceLevel,
  Direct3DTexture8_LockRect,
  Direct3DTexture8_UnlockRect,
  Direct3DTexture8_AddDirtyRect
};

HRESULT WINAPI Direct3DDevice8_CreateTexture(LPDIRECT3DDEVICE8 iface,
					     UINT Width,
					     UINT Height,
					     UINT Levels,
					     DWORD Usage,
					     D3DFORMAT Format,
					     D3DPOOL Pool,
					     IDirect3DTexture8** ppTexture)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  IDirect3DTexture8Impl *pTexture;
  UINT levels = Levels;
  DDSCAPS2 caps;
  unsigned u;
  HRESULT hr;

  TRACE("(%p}->(%d,%d,%d,%ld,0x%x,%d,%p)\n", iface, Width, Height,
	Levels, Usage, Format, Pool, ppTexture);

  if (!levels) {
    /* calculate levels */
    UINT Dim = (Width > Height) ? Width : Height;
    levels = 1;
    while (Dim > 1) {
      Dim = Dim / 2;
      levels++;
    }
  }

  pTexture = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(IDirect3DTexture8Impl) +
		       (levels-1) * sizeof(LPDIRECT3DSURFACE8));
  pTexture->lpVtbl = &D3DT8_Vtbl;
  pTexture->ref = 1;
  pTexture->parent = This;
  pTexture->Width  = Width;
  pTexture->Height = Height;
  pTexture->Levels = levels;
  pTexture->Usage  = Usage;
  pTexture->Format = Format;

  memset(&caps, 0, sizeof(caps));
  /* FIXME: handle caps bits */
  caps.dwCaps = DDSCAPS_TEXTURE;
  caps.dwCaps2 = 0;
  caps.dwCaps3 = 0;
  caps.dwCaps4 = 0;
  if (Usage & D3DUSAGE_RENDERTARGET) caps.dwCaps |= DDSCAPS_3DDEVICE;
  if (Usage & D3DUSAGE_WRITEONLY)    caps.dwCaps |= DDSCAPS_WRITEONLY;

  if (levels > 1) caps.dwCaps |= DDSCAPS_MIPMAP;

  for (u=0; u<levels; u++) {
    hr = Direct3DDevice8_CreateSurface(iface, (LPUNKNOWN)pTexture,
				       Width, Height, Format,
				       &caps, &pTexture->surf[u]);
    if (u) {
      Direct3DDevice8_AttachSurface(pTexture->surf[u-1],
				    pTexture->surf[u]);
    } else {
      caps.dwCaps2 |= DDSCAPS2_MIPMAPSUBLEVEL;
    }
    if (Width  > 1) Width  = Width / 2;
    if (Height > 1) Height = Height / 2;
  }

  TRACE("created %p\n", pTexture);
  *ppTexture = (LPDIRECT3DTEXTURE8)pTexture;
  
  /* add texture to linked list */
  if (!This->textures)
  {
    This->textures = pTexture;
  }
  else
  {
    This->textures->prev_device = pTexture;
    pTexture->next_device = This->textures;
    This->textures = pTexture;
  }

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_UpdateTexture(LPDIRECT3DTEXTURE8 iface,
                                             IDirect3DBaseTexture8 *pSourceTexture,
                                             IDirect3DBaseTexture8 *pDestinationTexture)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);

  FIXME("(%p)->(%p, %p): stub\n", iface, pSourceTexture, pDestinationTexture);
  return D3DERR_INVALIDCALL;
}

VOID WINAPI Direct3DTexture8_UpdateDirty(LPDIRECT3DTEXTURE8 iface, BOOL force)
{
  ICOM_THIS(IDirect3DTexture8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXBLT, 0, {1} };
  D3DHAL_DP2TEXBLT data;
  unsigned u;

  TRACE("(%p)->(%d)\n", iface, force);
  if (IsRectEmpty(&This->DirtyRect)) {
    if (!force) return;
    data.rSrc.left = 0;
    data.rSrc.top = 0;
    data.rSrc.right = This->Width;
    data.rSrc.bottom = This->Height;
  }
  else memcpy(&data.rSrc, &This->DirtyRect, sizeof(RECT));
  SetRectEmpty(&This->DirtyRect);

  data.dwDDSrcSurface = 0;
  data.pDest.x = 0;
  data.pDest.y = 0;
  data.dwFlags = 0;

  for (u=0; u<This->Levels; u++) {
    IDirect3DSurface8Impl* surf = (IDirect3DSurface8Impl*)This->surf[u];
    data.dwDDDestSurface = (DWORD)&surf->t;
    This->parent->emit_dp2(This->parent->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
    if (surf->t.gbl.wWidth > 1) {
      data.rSrc.left = data.rSrc.left / 2;
      data.rSrc.right = (data.rSrc.right+1) / 2;
    }
    if (surf->t.gbl.wHeight > 1) {
      data.rSrc.top = data.rSrc.top / 2;
      data.rSrc.bottom = (data.rSrc.bottom+1) / 2;
    }
  }
}
