/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

/* #define DUMP_TEXTURE_DATA */
/* unfortunately, GIMP don't seem to grok BITMAPV4HEADER (or BITMAPV5HEADER) yet */
#define DUMP_OLD_INFOHEADER

DEFAULT_DEBUG_CHANNEL(ddraw);

#define SURF_LCL(surf)  (&(surf)->lcl)
#define SURF_MORE(surf) (&(surf)->more)
#define SURF_GBL(surf)  (&(surf)->gbl)
#define TDATA(surf)     ((surf)->tdata)
#define TGLDEVICE(surf) (TDATA(surf)[0])
#define TGLFMT(surf)    (TDATA(surf)[1])
#define TGLTYPE(surf)   (TDATA(surf)[2])
#define TGLLEVEL(surf)  (TDATA(surf)[3])
#define TGLIFMT(surf)   (TDATA(surf)[4])

#define TEXTURE_ID(surf) (GET_LPDDRAWSURFACE_GBL_MORE(SURF_GBL(surf))->dwDriverReserved)

/* stuff we use that are still not in SGI's glext.h */
/* token values taken from http://oss.sgi.com/projects/ogl-sample/registry/ */
#ifndef GL_ARB_texture_border_clamp
#define GL_CLAMP_TO_BORDER_ARB	0x812D
#endif
#ifndef GL_ARB_texture_env_combine
#define GL_SUBTRACT_ARB		0x84E7
#endif
#ifndef GL_ARB_texture_env_dot3
#define GL_DOT3_RGB_ARB		0x86AE
#define GL_DOT3_RGBA_ARB	0x86AF
#endif
#ifndef GL_EXT_texture_env_dot3 /* ext 220 */
#define GL_DOT3_RGB_EXT		0x8740
#define GL_DOT3_RGBA_EXT	0x8741
#endif
#ifndef GL_IBM_mirrored_repeat /* ext 224 */
#define GL_MIRRORED_REPEAT_IBM	0x8370
#endif

#ifdef DUMP_TEXTURE_DATA
static void dump_texture(LPD3D_T texture)
{
    static unsigned count = 0;
    char buf[128];
    FILE* f;
    size_t size, y;
    BITMAPFILEHEADER fh;
#ifdef DUMP_OLD_INFOHEADER
    BITMAPINFOHEADER bh;
#else
    BITMAPV4HEADER bh;
#endif
    LPBYTE data;

    sprintf(buf, "texture_%d_%d.bmp", TEXTURE_ID(texture), ++count);

    size = SURF_GBL(texture)->u4.lPitch * SURF_GBL(texture)->wHeight;

    memset(&fh, 0, sizeof(fh));
    fh.bfType = 0x4d42; /* BM */
    fh.bfSize = sizeof(fh) + sizeof(bh) + size;
    fh.bfOffBits = sizeof(fh) + sizeof(bh);

    memset(&bh, 0, sizeof(bh));
#ifdef DUMP_OLD_INFOHEADER
    bh.biSize = sizeof(bh);
    bh.biWidth = SURF_GBL(texture)->wWidth;
    bh.biHeight = SURF_GBL(texture)->wHeight;
    bh.biPlanes = 1;
    bh.biBitCount = SURF_GBL(texture)->ddpfSurface.u1.dwRGBBitCount;
    bh.biCompression = BI_RGB;
    bh.biSizeImage = size;
    if (bh.biBitCount == 16 || bh.biBitCount == 32) {
	bh.biCompression = BI_BITFIELDS;
	fh.bfOffBits += 3 * sizeof(DWORD);
    }
#else
    bh.bV4Size = sizeof(bh);
    bh.bV4Width = SURF_GBL(texture)->wWidth;
    bh.bV4Height = SURF_GBL(texture)->wHeight;
    bh.bV4Planes = 1;
    bh.bV4BitCount = SURF_GBL(texture)->ddpfSurface.u1.dwRGBBitCount;
    bh.bV4Compression = BI_BITFIELDS;
    bh.bV4SizeImage = size;
    bh.bV4RedMask   = SURF_GBL(texture)->ddpfSurface.u2.dwRBitMask;
    bh.bV4GreenMask = SURF_GBL(texture)->ddpfSurface.u3.dwGBitMask;
    bh.bV4BlueMask  = SURF_GBL(texture)->ddpfSurface.u4.dwBBitMask;
    bh.bV4AlphaMask = SURF_GBL(texture)->ddpfSurface.u5.dwRGBAlphaBitMask;
    bh.bV4CSType = LCS_DEVICE_RGB;
#endif

    f = fopen(buf, "w+");
    fwrite(&fh, 1, sizeof(fh), f);
    fwrite(&bh, 1, sizeof(bh), f);

#ifdef DUMP_OLD_INFOHEADER
    if (bh.biBitCount == 16 || bh.biBitCount == 32) {
	fwrite(&SURF_GBL(texture)->ddpfSurface.u2.dwRBitMask, 1, sizeof(DWORD), f);
	fwrite(&SURF_GBL(texture)->ddpfSurface.u3.dwGBitMask, 1, sizeof(DWORD), f);
	fwrite(&SURF_GBL(texture)->ddpfSurface.u4.dwBBitMask, 1, sizeof(DWORD), f);
    }
#endif

    data = (LPVOID)SURF_GBL(texture)->fpVidMem;
    y = SURF_GBL(texture)->wHeight;
    while (y) {
	y--;
	fwrite(data + y*SURF_GBL(texture)->u4.lPitch, 1, SURF_GBL(texture)->u4.lPitch, f);
    }
    fclose(f);
}
#endif

static inline float GL_RSF(DWORD dwState)
{
    return *(float*)&dwState;
}

static inline void GL_RS2Col(GLfloat col[4], DWORD dwState)
{
    D3DCOLOR rcol = (D3DCOLOR)dwState;
    col[0] = RGBA_GETRED(rcol)   / 255.0;
    col[1] = RGBA_GETGREEN(rcol) / 255.0;
    col[2] = RGBA_GETBLUE(rcol)  / 255.0;
    col[3] = RGBA_GETALPHA(rcol) / 255.0;
}

/* Right now, this only supports the formats we enumerate. */
static BOOL GL_PixelFormat_To_TextureFormat(GL_D3D_priv *priv,
					    const DDPIXELFORMAT *pfmt,
					    GLenum *fmt,
					    GLint *ifmt,
					    GLenum *type)
{
    static const DWORD unsupported = DDPF_BUMPDUDV | DDPF_BUMPLUMINANCE
	| DDPF_PALETTEINDEXED1
	| DDPF_PALETTEINDEXED2 | DDPF_PALETTEINDEXED4 | DDPF_PALETTEINDEXED8
	| DDPF_PALETTEINDEXEDTO8 | DDPF_RGBTOYUV | DDPF_STENCILBUFFER
	| DDPF_YUV | DDPF_ZBUFFER | DDPF_ZPIXELS;

    /* And ALPHAPREMULT might cause trouble. */

    TRACE("surface format %08lx, %ld\n", pfmt->dwFlags, pfmt->u1.dwRGBBitCount);
    TRACE("masks: %08lx, %08lx, %08lx, %08lx\n", pfmt->u2.dwRBitMask, pfmt->u3.dwGBitMask, pfmt->u4.dwBBitMask, pfmt->u5.dwRGBAlphaBitMask);

    if (pfmt->dwFlags & unsupported)
	return FALSE;

    if (pfmt->dwFlags & DDPF_FOURCC)
    {
	DD_GL_ext ext;
	TRACE("fourcc: %4.4s\n", (const char*)&pfmt->dwFourCC);
	if (!HAVE_GL_EXT(priv, ARB_texture_compression))
	    return FALSE;
	ext = 0;
	switch (pfmt->dwFourCC) {
	case 0x31545844: /* DXT1 */
	    if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
		*fmt = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
	    else
		*fmt = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	case 0x33545844: /* DXT3 */
	    *fmt = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	case 0x35545844: /* DXT5 */
	    *fmt = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	}
	if (!ext) {
	    FIXME("unsupported texture format %4.4s\n", (const char*)&pfmt->dwFourCC);
	    return FALSE;
	}
	if (!HAVE_GL_EXT(priv, ext))
	    return FALSE;
	*ifmt = *fmt;
	*type = 0;
	return TRUE;
    }

    if (pfmt->dwFlags & DDPF_COMPRESSED)
	return FALSE;

    if (pfmt->dwFlags & DDPF_RGB)
    {
	if (pfmt->dwFlags & (DDPF_LUMINANCE|DDPF_ALPHA))
	    return FALSE;

	if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
	{
	    *ifmt = GL_RGBA;
	    *fmt = GL_BGRA;

	    if (pfmt->u1.dwRGBBitCount == 32)
	    {
		*type = GL_UNSIGNED_BYTE;
		return TRUE;
	    }
	    else if (pfmt->u1.dwRGBBitCount == 16)
	    {
		if (pfmt->u5.dwRGBAlphaBitMask == 0x8000)
		    *type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
		else
		    *type = GL_UNSIGNED_SHORT_4_4_4_4_REV;

		return TRUE;
	    }
	    else
		return FALSE;
	}
	else
	{
	    *ifmt = GL_RGB;

	    if (pfmt->u1.dwRGBBitCount == 32)
	    {
		*fmt = GL_BGRA;
		*type = GL_UNSIGNED_BYTE;
		return TRUE;
	    }
	    else if (pfmt->u1.dwRGBBitCount == 16)
	    {
		if (pfmt->u2.dwRBitMask & 0x8000)
		{
		    *fmt = GL_RGB;
		    *type = GL_UNSIGNED_SHORT_5_6_5;
		}
		else
		{
		    *fmt = GL_BGRA;
		    *type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
		}

		return TRUE;
	    }
	    else
		return FALSE;
	}
    }

    if (pfmt->dwFlags & DDPF_LUMINANCE)
    {
	if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
	{
	    *fmt = GL_LUMINANCE_ALPHA;
	    *ifmt = GL_LUMINANCE_ALPHA;
	    *type = GL_UNSIGNED_BYTE;
	}
	else
	{
	    *fmt = GL_LUMINANCE;
	    *ifmt = GL_LUMINANCE;
	    *type = GL_UNSIGNED_BYTE;
	}

	return TRUE;
    }

    return FALSE;
}

#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
static unsigned int GL_count_levels(unsigned int width, unsigned int height)
{
    unsigned int x = max(width, height);
    unsigned int levels = 0;

    while (x > 1)
    {
	x /= 2;
	levels++;
    }

    return levels;
}
#endif

static LPD3D_T GL_next_mipmap(LPD3D_T texture)
{
    LPATTACHLIST next = SURF_LCL(texture)->lpAttachList;
    while (next) {
	LPD3D_T surf = (LPD3D_T)next->lpAttached;
	if ((SURF_LCL(surf)->ddsCaps.dwCaps & DDSCAPS_TEXTURE) &&
	    (SURF_MORE(surf)->ddsCapsEx.dwCaps2 & DDSCAPS2_MIPMAPSUBLEVEL))
	    return surf;
	next = next->lpLink;
    }
    return NULL;
}

static HRESULT GL_PrepTexture(GL_D3D_priv *priv, LPD3D_T texture)
{
    LPDDPIXELFORMAT pfmt = &(SURF_GBL(texture)->ddpfSurface);
    GLint ifmt;
    GLenum fmt, type;
    LPD3D_T mipmap;
    int level;

    if (!TEXTURE_ID(texture)) {
	LPDDRAWI_DDRAWSURFACE_LCL local = SURF_LCL(texture);
	DDHAL_CREATESURFACEDATA data;
	HRESULT hr;

	/* store which 3D context texture belongs to */
	local->dwReserved1 = priv->d3dp->dwhContext;

	/* call upon the HAL to allocate a texture ID */
	data.lpDD = SURF_GBL(texture)->u3.lpDD;
	data.lpDDSurfaceDesc = (LPDDSURFACEDESC)(SURF_MORE(texture)->pCreatedDDSurfaceDesc2);
	data.lplpSList = &local;
	data.dwSCnt = 1;
	data.ddRVal = 0;
	data.CreateSurface = priv->dd_gbl->lpDDCBtmp->HALDD.CreateSurface;
	hr = data.CreateSurface(&data);

	if (!TEXTURE_ID(texture)) {
	    ERR("can't allocate texture ID\n");
	    return DDERR_GENERIC;
	}
    }

    if (!GL_PixelFormat_To_TextureFormat(priv, pfmt, &fmt, &ifmt, &type))
    {
	FIXME("unhandled format %lx, %ld\n", pfmt->dwFlags,
	      pfmt->u1.dwRGBBitCount);
	return DDERR_INVALIDPARAMS;
    }
    TRACE("ID %d, fmt %X, ifmt %X, type %X\n", TEXTURE_ID(texture), fmt, ifmt, type);

    glPushAttrib(GL_TEXTURE_BIT); /* save the current texture bindings */

    glBindTexture(GL_TEXTURE_2D, (GLuint)TEXTURE_ID(texture));

#ifdef DUMP_TEXTURE_DATA
    dump_texture(texture);
#endif

    mipmap = texture;
    level = 0;

    while (1)
    {
	LPD3D_T next;

	TRACE("loading level %d, %dx%d, ptr %p\n", level,
	      SURF_GBL(mipmap)->wWidth, SURF_GBL(mipmap)->wHeight,
	      (LPVOID)SURF_GBL(mipmap)->fpVidMem);

	if (type)
	    glTexImage2D(GL_TEXTURE_2D, level, ifmt,
			 SURF_GBL(mipmap)->wWidth,
			 SURF_GBL(mipmap)->wHeight,
			 0, fmt, type,
			 (LPVOID)SURF_GBL(mipmap)->fpVidMem);
	else
	    GLD(priv)->glCompressedTexImage2DARB(GL_TEXTURE_2D, level, ifmt,
						 SURF_GBL(mipmap)->wWidth,
						 SURF_GBL(mipmap)->wHeight,
						 0, SURF_GBL(mipmap)->u4.dwLinearSize,
						 (LPVOID)SURF_GBL(mipmap)->fpVidMem);

	TGLDEVICE(mipmap) = (DWORD)priv;
	TGLFMT(mipmap) = fmt;
	TGLIFMT(mipmap) = ifmt;
	TGLTYPE(mipmap) = type;
	TGLLEVEL(mipmap) = level;

	level++;

	/* Get the next mipmap. */
	if (!(next = GL_next_mipmap(mipmap)))
	{
	    TRACE("provided mipmap levels: %d\n", level);
	    level--; /* level is the max (smallest) mipmap level */

	    /* Fill in the remaining mipmaps as a workaround for drivers that
	     * do not support GL_TEXTURE_MAX_LEVEL.
	     *
	     * XXX Should we do this for compressed textures too?
	     *
	     * Note that old Mesa libGLU may not have this function.
	     * Get a recent version that includes the SGI libGLU.
	     */
#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
	    if (type
		&& (SURF_GBL(mipmap)->wWidth > 1
		    || SURF_GBL(mipmap)->wHeight > 1))
	    {
		DWORD max_level = level
		    + GL_count_levels(SURF_GBL(mipmap)->wWidth,
				      SURF_GBL(mipmap)->wHeight);

		TRACE("generating mipmap levels %d-%ld\n", level+1, max_level);
		gluBuild2DMipmapLevels(GL_TEXTURE_2D, ifmt,
				       SURF_GBL(mipmap)->wWidth,
				       SURF_GBL(mipmap)->wHeight,
				       fmt, type, level, level+1, max_level,
				       (LPVOID)SURF_GBL(mipmap)->fpVidMem);
	    }
#endif
	    break;
	}

	mipmap = next;
    }

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, level);

    glPopAttrib();

    TRACE("initialized GL texture %d\n", TEXTURE_ID(texture));

    return D3D_OK;
}

#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
static BOOL
D3D_GL_is_last_mipmap(LPD3D_T texture)
{
    return !GL_next_mipmap(texture);
}
#endif

HRESULT D3D_GL_texblt(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2TEXBLT dat = (LPD3DHAL_DP2TEXBLT)data1;
    LPD3D_T srctex = (LPD3D_T)dat->dwDDSrcSurface;
    LPD3D_T texture = (LPD3D_T)dat->dwDDDestSurface;
    GLint oldtex;

    TRACE("(%p): src=%p, dst=%p\n", priv, srctex, texture);

    /* PreLoad? */
    if (!texture) {
	if (!TGLDEVICE(srctex))
	    return GL_PrepTexture(priv, srctex);
	return D3D_OK;
    }

    /* Texture previously been loaded into this context? */
    if (TGLDEVICE(texture) != (DWORD)priv) return D3D_OK;

    assert(srctex == NULL);

    /* OK, perform requested update */
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &oldtex);
    glBindTexture(GL_TEXTURE_2D, TEXTURE_ID(texture));
    if (TGLTYPE(texture))
    {
	glTexSubImage2D(GL_TEXTURE_2D, TGLLEVEL(texture),
			dat->rSrc.left, dat->rSrc.top,
			dat->rSrc.right - dat->rSrc.left, dat->rSrc.bottom - dat->rSrc.top,
			TGLFMT(texture), TGLTYPE(texture), (LPVOID)SURF_GBL(texture)->fpVidMem);

	/* Despite GL_TEXTURE_MAX_LEVEL, we actually load all the mipmap
	 * levels to support buggy hardware.
	 * If the smallest mipmap level is being reloaded, then we need to
	 * regenerate the mipmap levels from it. */
#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
	if (D3D_GL_is_last_mipmap(texture)
	    && (SURF_GBL(texture)->wWidth > 1
		|| SURF_GBL(texture)->wHeight > 1))
	
	{
	    DWORD max_level = TGLLEVEL(texture);
		+ GL_count_levels(SURF_GBL(texture)->wWidth,
				  SURF_GBL(texture)->wHeight);

	    gluBuild2DMipmapLevels(GL_TEXTURE_2D, TGLIFMT(texture),
				   SURF_GBL(texture)->wWidth,
				   SURF_GBL(texture)->wHeight,
				   TGLFMT(texture), TGLTYPE(texture),
				   TGLLEVEL(texture), TGLLEVEL(texture)+1,
				   max_level, (LPVOID)SURF_GBL(texture)->fpVidMem);
	}
#endif

    }
    else
	FIXME("update compressed texture\n");
    glBindTexture(GL_TEXTURE_2D, oldtex);

    TRACE("updated GL texture %d\n", TEXTURE_ID(texture));

    return D3D_OK;
}

typedef const struct {
    DD_GL_ext ext;
    DWORD dwFlags;
    GLint scale, cmode, amode, src0, src1, src2;
} GL_TEXOP;

#define E_GL 0
#define E_EA EXT_texture_env_add
#define E_EC EXT_texture_env_combine
#define E_AC ARB_texture_env_combine
#define E_E3 EXT_texture_env_dot3
#define E_A3 ARB_texture_env_dot3

/* If a SOURCE bit is set, that src value tells that a D3D texop argument should be
 * mapped and substituted to the corresponding GL source. If not, the src value is
 * considered a GL token, and is used directly (with a default operand), if non-zero.
 * For instance, for color ops, if dwFlags = GLTF_SOURCE0, src0 = 0, will mean that
 * D3DTSS_COLORARG1 will be mapped into GL_SOURCE0_RGB_EXT. Similarly, src0 = 1 will
 * map D3DTSS_COLORARG2 into GL_SOURCE0_RGB_EXT instead. */
#define GLTF_SOURCE0	0x0001
#define GLTF_SOURCE1	0x0002
#define GLTF_SOURCE2	0x0004
#define GLTF_SOURCE3	0x0008
#define GLTF_SOURCEA	GLTF_SOURCE0|GLTF_SOURCE1
#define GLTF_SOURCEB	GLTF_SOURCE1|GLTF_SOURCE2
#define GLTF_SOURCEC	GLTF_SOURCE0|GLTF_SOURCE1|GLTF_SOURCE2
#define GLTF_COMPL0	0x0010
#define GLTF_COMPL1	0x0020
#define GLTF_COMPL2	0x0040
#define GLTF_COMPL3	0x0080
/* Modifiers, not implemented or well-defined yet. */
#define GLTF_SA		0x0100	/* Special alpha treatment */
#define GLTF_C		0x0200	/* Use a texture environment constant */
#define GLTF_NC		0x0400	/* Don't use combiners */

/* If you implement anything, also try to update caps bits in ddraw/gl.c */

/* this table is indexed by the table below */
static GL_TEXOP GL_tex_op_data[] = {
/* supported by OpenGL 1.2 */
    /* SELECTARG1 */			{E_GL, GLTF_SOURCE0, 1, GL_REPLACE, 0, 0},
    /* SELECTARG2 */			{E_GL, GLTF_SOURCE0, 1, GL_REPLACE, 0, 1},
    /* MODULATE */			{E_GL, GLTF_SOURCEA, 1, GL_MODULATE, 0, 0, 1},
/* supported by ARB_texture_env_add */
    /* ADD */				{E_EA, GLTF_SOURCEA, 1, GL_ADD, 0, 0, 1},
/* supported by ARB_texture_env_combine */
    /* ADDSMOOTH */			{E_AC, GLTF_SOURCEB|GLTF_C, 1, GL_INTERPOLATE_EXT, 0, GL_CONSTANT_EXT, 1, 0},
    /* ADDSMOOTH */			{E_GL, GLTF_SOURCEA|GLTF_C|GLTF_NC, 1, GL_BLEND, 0, 0, 1},
    /* SUBTRACT */			{E_AC, GLTF_SOURCEA, 1, GL_SUBTRACT_ARB, 0, 0, 1},
/* supported by EXT_texture_env_combine */
    /* SUBTRACT */			{E_EC, GLTF_SOURCEA|GLTF_COMPL1, 1, GL_ADD, 0, 0, 1},
    /* MODULATE2X */			{E_EC, GLTF_SOURCEA, 2, GL_MODULATE, 0, 0, 1},
    /* MODULATE4X */			{E_EC, GLTF_SOURCEA, 4, GL_MODULATE, 0, 0, 1},
    /* ADDSIGNED */			{E_EC, GLTF_SOURCEA, 1, GL_ADD_SIGNED_EXT, 0, 0, 1},
    /* ADDSIGNED2X */			{E_EC, GLTF_SOURCEA, 2, GL_ADD_SIGNED_EXT, 0, 0, 1},
    /* BLENDDIFFUSEALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_PRIMARY_COLOR_EXT},
    /* BLENDTEXTUREALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_TEXTURE},
    /* BLENDTEXTUREALPHA */		{E_GL, GLTF_SOURCEA, 1, GL_DECAL, 0, 0, 1},
    /* BLENDFACTORALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_CONSTANT_EXT},
    /* BLENDCURRENTALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_PREVIOUS_EXT},
    /* LERP */				{E_EC, GLTF_SOURCEC, 1, GL_INTERPOLATE_EXT, 0, 1, 2, 0},
/* supported by ARB_texture_env_dot3 or EXT_texture_env_dot3 (incompatible extensions) */
    /* DOTPRODUCT3 */			{E_A3, GLTF_SOURCEA|GLTF_SA, 1, GL_DOT3_RGB_ARB, GL_DOT3_RGBA_ARB, 0, 1},
    /* DOTPRODUCT3 */			{E_E3, GLTF_SOURCEA|GLTF_SA, 1, GL_DOT3_RGB_EXT, GL_DOT3_RGBA_EXT, 0, 1}
};

static struct {
    int c;
    GL_TEXOP *op;
} GL_tex_op[] = {
    {0,NULL}, {0,NULL},      /* DISABLE */
    {1,&GL_tex_op_data[ 0]}, /* SELECTARG1 */
    {1,&GL_tex_op_data[ 1]}, /* SELECTARG2 */
    {1,&GL_tex_op_data[ 2]}, /* MODULATE */
    {1,&GL_tex_op_data[ 8]}, /* MODULATE2X */
    {1,&GL_tex_op_data[ 9]}, /* MODULATE4X */
    {1,&GL_tex_op_data[ 3]}, /* ADD */
    {1,&GL_tex_op_data[10]}, /* ADDSIGNED */
    {1,&GL_tex_op_data[11]}, /* ADDSIGNED2X */
    {2,&GL_tex_op_data[ 6]}, /* SUBTRACT */
    {2,&GL_tex_op_data[ 4]}, /* ADDSMOOTH */
    {1,&GL_tex_op_data[12]}, /* BLENDDIFFUSEALPHA */
    {2,&GL_tex_op_data[13]}, /* BLENDTEXTUREALPHA */
    {1,&GL_tex_op_data[15]}, /* BLENDFACTORALPHA */
    {0,NULL},                /* BLENDTEXTUREALPHAPM */
    {1,&GL_tex_op_data[16]}, /* BLENDCURRENTALPHA */
    {0,NULL},                /* PREMODULATE */
    {0,NULL},                /* MODULATEALPHA_ADDCOLOR */
    {0,NULL},                /* MODULATECOLOR_ADDALPHA */
    {0,NULL},                /* MODULATEINVALPHA_ADDCOLOR */
    {0,NULL},                /* MODULATEINVCOLOR_ADDALPHA */
    {0,NULL},                /* BUMPENVMAP */
    {0,NULL},                /* BUMPENVMAPLUMINANCE */
    {2,&GL_tex_op_data[18]}, /* DOTPRODUCT3 */
    {0,NULL},                /* MULTIPLYADD */
    {1,&GL_tex_op_data[17]}  /* LERP */
};

static const GLint GL_D3DTA[] = {
    GL_PRIMARY_COLOR_EXT, /* diffuse */
    GL_PREVIOUS_EXT,
    GL_TEXTURE,
    GL_CONSTANT_EXT,
    GL_PRIMARY_COLOR_EXT /* specular, not supported by GL yet */
};

static const GLenum GL_mag[] = {
    0,
    GL_NEAREST,	/* POINT */
    GL_LINEAR,	/* LINEAR */
    GL_LINEAR,	/* FLATCUBIC */
    GL_LINEAR,	/* GAUSSIANCUBIC */
    GL_LINEAR	/* ANISOTROPIC */
};

static const GLenum GL_mip_min[4][4] = {
/* min: POINT                      LINEAR                    ANISOTROPIC */
    {0, 0,                         0,                        0},			/* mip: */
    {0, GL_NEAREST,                GL_LINEAR,                GL_LINEAR},		/* NONE */
    {0, GL_NEAREST_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST},	/* POINT */
    {0, GL_NEAREST_MIPMAP_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,  GL_LINEAR_MIPMAP_LINEAR}	/* LINEAR */
};

static const GLenum GL_tex_addr[] = {
    0,
    GL_REPEAT,			/* WRAP */
    GL_MIRRORED_REPEAT_IBM,	/* MIRROR */
    GL_CLAMP_TO_EDGE,		/* CLAMP */
    GL_CLAMP_TO_BORDER_ARB	/* BORDER */
};

/* Returns whether we can use a particular extension */
static inline BOOL D3D_GL_have_tex_ext(GL_D3D_priv* priv, DD_GL_ext ext)
{
    if (ext == E_GL) return TRUE;

    if (HAVE_GL_EXT(priv, ext)) return TRUE;

    /* In case there are GL implementations that report ARB but not EXT */
    if (ext == EXT_texture_env_add) return HAVE_GL_EXT(priv, ARB_texture_env_add);
    if (ext == EXT_texture_env_combine) return HAVE_GL_EXT(priv, ARB_texture_env_combine);

    return FALSE;
}

/* Find the most appropriate texture operation descriptor from GL_tex_op table
 * Returns NULL if no available GL extension can support the requested operation
 */
static inline GL_TEXOP* D3D_GL_get_tex_op(GL_D3D_priv* priv, DWORD op)
{
    int c;
    GL_TEXOP* gop;
    c = GL_tex_op[op].c;
    gop = GL_tex_op[op].op;
    while (c) {
	if (D3D_GL_have_tex_ext(priv, gop->ext)) return gop;
	gop++;
	c--;
    }
    return NULL;
}

#define OPERAND_OFS (GL_OPERAND0_RGB_EXT - GL_SOURCE0_RGB_EXT)

/* Compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter from D3D tex-op argument */
static inline GLenum GL_get_alpha_opn(DWORD dwValue) {
    if (dwValue & D3DTA_COMPLEMENT) return GL_ONE_MINUS_SRC_ALPHA;
    return GL_SRC_ALPHA;
}

static inline GLenum GL_get_color_opn(DWORD dwValue) {
    if (dwValue & D3DTA_ALPHAREPLICATE) return GL_get_alpha_opn(dwValue);
    if (dwValue & D3DTA_COMPLEMENT) return GL_ONE_MINUS_SRC_COLOR;
    return GL_SRC_COLOR;
}

static inline DWORD GL_get_compl(GL_TEXOP* gop, DWORD flag)
{
    if (gop->dwFlags & flag) return D3DTA_COMPLEMENT;
    return 0;
}

/* Template for inline functions to be given to COND_TEX_ENV.
 * One such function is instantiated for every possible GL tex-op source.
 * Template takes:
 *  name = name of function
 *  flag,src = GL tex-op source (see GLTF flags above)
 *  arg1,arg2 = D3D argument fields
 *  opn_func = function to compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter from D3D argument
 *  opn_def = default GL_OPERANDn_{RGB|ALPHA}_EXT if D3D arguments are not used
 * Instantiated function takes:
 *  dwStage = texture stage to take any tex-op arguments from
 *  gop = texture operation descriptor
 * Function returns:
 *  src = parameter for GL_SOURCEn_{RGB|ALPHA}_EXT
 *  opn = parameter for GL_OPERANDn_{RGB|ALPHA}_EXT
 */
#define MAKE_TEX_SRC(name,flag,src,arg1,arg2,arg0,opn_func,opn_def) \
static inline void name(LPD3D_P d3dp, DWORD dwStage, GL_TEXOP* gop, GLenum* src, GLenum* opn) \
{ \
    if (gop->dwFlags & (flag)) { \
	DWORD dwValue = ((gop->src == 2) ? d3dp->texstagestate[dwStage][D3DTSS_##arg0] : \
			 (gop->src == 1) ? d3dp->texstagestate[dwStage][D3DTSS_##arg2] : \
			 (gop->src == 0) ? d3dp->texstagestate[dwStage][D3DTSS_##arg1] : 0) \
			^ GL_get_compl(gop, (flag) << 4); \
	*src = GL_D3DTA[dwValue & D3DTA_SELECTMASK]; \
	*opn = opn_func(dwValue); \
    } else { \
	*src = gop->src; \
	*opn = opn_def; \
    } \
}

MAKE_TEX_SRC(GL_get_color_src0,GLTF_SOURCE0,src0,COLORARG1,COLORARG2,COLORARG0,GL_get_color_opn,GL_SRC_COLOR)
MAKE_TEX_SRC(GL_get_color_src1,GLTF_SOURCE1,src1,COLORARG1,COLORARG2,COLORARG0,GL_get_color_opn,GL_SRC_COLOR)
MAKE_TEX_SRC(GL_get_color_src2,GLTF_SOURCE2,src2,COLORARG1,COLORARG2,COLORARG0,GL_get_color_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src0,GLTF_SOURCE0,src0,ALPHAARG1,ALPHAARG2,ALPHAARG0,GL_get_alpha_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src1,GLTF_SOURCE1,src1,ALPHAARG1,ALPHAARG2,ALPHAARG0,GL_get_alpha_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src2,GLTF_SOURCE2,src2,ALPHAARG1,ALPHAARG2,ALPHAARG0,GL_get_alpha_opn,GL_SRC_ALPHA)

#undef MAKE_TEX_SRC

/* Template for inline functions to be given to COND_TEX_ARG.
 * One such function is instantiated for color, and one for alpha.
 * Template takes:
 *  name = name of function
 *  suff = GL tex-op source suffix (RGB or ALPHA)
 * Instantiated function takes:
 *  gop = texture operation descriptor
 *  no = D3D argument number (0 or 1, see GLTF flags above)
 * Function returns:
 *  which GL_SOURCEn_{RGB|ALPHA}_EXT parameter the D3D argument maps to.
 */
#define MAKE_TEX_SRC(name,suff) \
static inline GLenum name(GL_TEXOP* gop, int no) { \
    if ((gop->dwFlags & GLTF_SOURCE0) && gop->src0 == no) return GL_SOURCE0_##suff; \
    if ((gop->dwFlags & GLTF_SOURCE1) && gop->src1 == no) return GL_SOURCE1_##suff; \
    if ((gop->dwFlags & GLTF_SOURCE2) && gop->src2 == no) return GL_SOURCE2_##suff; \
    return 0; \
}

MAKE_TEX_SRC(GL_get_color_src,RGB_EXT)
MAKE_TEX_SRC(GL_get_alpha_src,ALPHA_EXT)

#undef MAKE_TEX_SRC

/* Conditional Texture Environment.
 * Sets GL tex-op parameters as required by the (newly set) texture operation.
 * Assumes:
 *  dwStage = texture stage
 * Takes:
 *  env = GL tex-op source to set
 *  gop = texture operation descriptor
 *  func = function to compute GL tex-op parameters
 */
#define COND_TEX_ENV(env, gop, func) \
    do { GLenum src, opn; \
	 func(d3dp, dwStage, gop, &src, &opn); \
	 TRACE(" arg %04x %04x %04x\n", env, src, opn); \
	 if (src) { glTexEnvi(GL_TEXTURE_ENV, env, src); \
	  glTexEnvi(GL_TEXTURE_ENV, env + OPERAND_OFS, opn); } } while (0)

/* Conditional Texture Argument.
 * Changes GL tex-op parameters to reflect a change in D3D tex-op arguments.
 * Assumes:
 *  dwStage = texture stage
 *  dwState = D3D tex-op argument that has changed
 * Takes:
 *  base = D3D tex-op argument base (to compute D3D argument number from)
 *  gop = texture operation descriptor
 *  env_func = function to compute which GL tex-op source to change
 *  opn_func = function to compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter
 */
#define COND_TEX_ARG(base, gop, env_func, opn_func) \
    do { DWORD dbase = ((dwState - base) > 1) ? 2 : (dwState - base); \
	 DWORD dwValue = d3dp->texstagestate[dwStage][dwState] \
			 ^ GL_get_compl(gop, 0x10 << dbase); \
	 GLenum env = env_func(gop, dbase), \
		src = GL_D3DTA[dwValue & D3DTA_SELECTMASK], \
		opn = opn_func(dwValue); \
	 TRACE(" arg %04x %04x %04x\n", env, src, opn); \
	 if (env) { glTexEnvi(GL_TEXTURE_ENV, env, src); \
	  glTexEnvi(GL_TEXTURE_ENV, env + OPERAND_OFS, opn); } } while (0)

static inline BOOL GL_IsStageUsed(GL_D3D_priv* priv, DWORD dwStage)
{
    LPD3D_P d3dp = priv->d3dp;
    if (!d3dp->texstagestate[dwStage][D3DTSS_TEXTUREMAP]) return FALSE;
    if (!D3D_GL_have_tex_ext(priv, EXT_texture_env_combine)) {
	DWORD arg = D3DTA_TEXTURE;
	/* if the app requests D3DTA_CURRENT (or D3DTA_DIFFUSE in stage 0) in
	 * a SELECTARG operation, then disabling the texture unit is the only
	 * way to get the desired pass-through operation with standard OpenGL */
	if (d3dp->texstagestate[dwStage][D3DTSS_COLOROP] == D3DTOP_SELECTARG1)
	    arg = d3dp->texstagestate[dwStage][D3DTSS_COLORARG1];
	if (d3dp->texstagestate[dwStage][D3DTSS_COLOROP] == D3DTOP_SELECTARG2)
	    arg = d3dp->texstagestate[dwStage][D3DTSS_COLORARG2];
	if (arg != D3DTA_TEXTURE) return FALSE;
    }
    return TRUE;
}

static inline BOOL GL_IsStageEnabled(GL_D3D_priv* priv, DWORD dwStage)
{
    LPD3D_P d3dp = priv->d3dp;
    unsigned u;
    for (u=0; u<=dwStage; u++)
	if (d3dp->texstagestate[u][D3DTSS_COLOROP] <= D3DTOP_DISABLE)
	    return FALSE;
    return GL_IsStageUsed(priv, dwStage);
}

static inline void GL_StageEnabled(GL_D3D_priv* priv, DWORD dwStage)
{
    LPD3D_P d3dp = priv->d3dp;
    unsigned u;
    if (GL_IsStageUsed(priv, dwStage))
	glEnable(GL_TEXTURE_2D);
    else
	glDisable(GL_TEXTURE_2D);
    for (u=dwStage+1; u<priv->lpCaps->wMaxTextureBlendStages &&
		      d3dp->texstagestate[u][D3DTSS_COLOROP]>D3DTOP_DISABLE; u++) {
	if (GL_IsStageUsed(priv, u)) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glEnable(GL_TEXTURE_2D);
	}
    }
}

static inline void GL_StageDisabled(GL_D3D_priv* priv, DWORD dwStage)
{
    LPD3D_P d3dp = priv->d3dp;
    unsigned u;
    if (GL_IsStageUsed(priv, dwStage))
	glDisable(GL_TEXTURE_2D);
    for (u=dwStage+1; u<priv->lpCaps->wMaxTextureBlendStages &&
		      d3dp->texstagestate[u][D3DTSS_COLOROP]>D3DTOP_DISABLE; u++) {
	if (GL_IsStageUsed(priv, u)) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glDisable(GL_TEXTURE_2D);
	}
    }
}

static void D3D_GL_update_texture_factor(GL_D3D_priv* priv, GL_TEXOP* cop, GL_TEXOP* aop)
{
    static const GLfloat one4f[4] = {1.0, 1.0, 1.0, 1.0};
    const GLfloat *factor = priv->texture_factor;
    if ((cop && (cop->dwFlags & GLTF_C)) || (aop && (aop->dwFlags & GLTF_C)))
	factor = one4f;
    TRACE(" factor (%f,%f,%f,%f)\n", factor[0], factor[1], factor[2], factor[3]);
    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, factor);
}

static void D3D_GL_set_tex_op(GL_D3D_priv* priv, DWORD dwStage, D3DTEXTURESTAGESTATETYPE dwState)
{
    LPD3D_P d3dp = priv->d3dp;
    DWORD cstate = d3dp->texstagestate[dwStage][D3DTSS_COLOROP];
    DWORD astate = d3dp->texstagestate[dwStage][D3DTSS_ALPHAOP];
    GL_TEXOP *cop = D3D_GL_get_tex_op(priv, cstate);
    GL_TEXOP *aop = D3D_GL_get_tex_op(priv, astate);

    if (D3D_GL_have_tex_ext(priv, EXT_texture_env_combine)) {
	switch (dwState) {
	case D3DTSS_COLOROP: {
		GLint cmode;
		if (cstate == D3DTOP_DISABLE) {
		    TRACE(" stage disable\n");
		    GL_StageDisabled(priv, dwStage);
		    return;
		}
		if (!cop) {
		    TRACE(" unsupported\n");
		    D3D_GL_update_texture_factor(priv, cop, aop);
		    return;
		}
		if (cop->dwFlags & GLTF_NC) goto no_combiner;
		cmode = cop->cmode;
		if (cop->dwFlags & GLTF_SA) {
		    if (astate == cstate)
			cmode = cop->amode;
		}
		TRACE(" color mode %04x\n", cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
		glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, cop->scale);
		COND_TEX_ENV(GL_SOURCE0_RGB_EXT, cop, GL_get_color_src0);
		COND_TEX_ENV(GL_SOURCE1_RGB_EXT, cop, GL_get_color_src1);
		COND_TEX_ENV(GL_SOURCE2_RGB_EXT, cop, GL_get_color_src2);
		D3D_GL_update_texture_factor(priv, cop, aop);
		GL_StageEnabled(priv, dwStage);
	    }
	    break;
	case D3DTSS_ALPHAOP: {
		GLint amode;
		if (!aop) {
		    if ((astate == D3DTOP_DISABLE) && (cstate == D3DTOP_DISABLE)) {
			TRACE(" stage disable\n");
			return;
		    }
		    TRACE(" unsupported\n");
		    D3D_GL_update_texture_factor(priv, cop, aop);
		    return;
		}
		amode = aop->cmode;
		if (cop && (cop->dwFlags & GLTF_SA)) {
		    if (cstate == astate) {
			TRACE(" color mode %04x\n", cop->amode);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, cop->amode);
			amode = cop->amode;
		    }
		    else {
			TRACE(" color mode %04x\n", cop->cmode);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, cop->cmode);
		    }
		}
		TRACE(" alpha mode %04x\n", amode);
		glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, amode);
		glTexEnvi(GL_TEXTURE_ENV, GL_ALPHA_SCALE, aop->scale);
		COND_TEX_ENV(GL_SOURCE0_ALPHA_EXT, aop, GL_get_alpha_src0);
		COND_TEX_ENV(GL_SOURCE1_ALPHA_EXT, aop, GL_get_alpha_src1);
		COND_TEX_ENV(GL_SOURCE2_ALPHA_EXT, aop, GL_get_alpha_src2);
		D3D_GL_update_texture_factor(priv, cop, aop);
	    }
	    break;
	case D3DTSS_COLORARG1:
	case D3DTSS_COLORARG2:
	case D3DTSS_COLORARG0:
	    {
		if (!cop) return;
		COND_TEX_ARG(D3DTSS_COLORARG1, cop, GL_get_color_src, GL_get_color_opn);
	    }
	    break;
	case D3DTSS_ALPHAARG1:
	case D3DTSS_ALPHAARG2:
	case D3DTSS_ALPHAARG0:
	    {
		if (!aop) return;
		COND_TEX_ARG(D3DTSS_ALPHAARG1, aop, GL_get_alpha_src, GL_get_alpha_opn);
	    }
	    break;
	default:
	}
	return;
    }
    if (dwState == D3DTSS_COLOROP) {
	if (cstate == D3DTOP_DISABLE) {
	    GL_StageDisabled(priv, dwStage);
	    return;
	}
	if (!cop) {
	    TRACE(" unsupported\n");
	    return;
	}
no_combiner:
	TRACE(" mode %04x\n", cop->cmode);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, cop->cmode);
	D3D_GL_update_texture_factor(priv, cop, aop);
	GL_StageEnabled(priv, dwStage);
    }
}

#undef COND_TEX_ARG
#undef COND_TEX_ENV

static HRESULT D3D_GL_validate_tex_stage(GL_D3D_priv* priv, DWORD dwStage)
{
    LPD3D_P d3dp = priv->d3dp;
    GL_TEXOP *cop = D3D_GL_get_tex_op(priv, d3dp->texstagestate[dwStage][D3DTSS_COLOROP]);
    GL_TEXOP *aop = D3D_GL_get_tex_op(priv, d3dp->texstagestate[dwStage][D3DTSS_ALPHAOP]);
    GLenum csrc0, csrc1, csrc2;
    GLenum copn0, copn1, copn2;
    GLenum asrc0, asrc1, asrc2;
    GLenum aopn0, aopn1, aopn2;

    if (!cop) return D3DERR_UNSUPPORTEDCOLOROPERATION;
    if (!aop) {
	if (d3dp->texstagestate[dwStage][D3DTSS_ALPHAOP] != D3DTOP_DISABLE)
	    return D3DERR_UNSUPPORTEDALPHAOPERATION;
        FIXME("trying to validate disabled alpha operation\n");   
	/* FIXME: maybe some apps might expect this to succeed anyway? */
	return D3DERR_UNSUPPORTEDALPHAOPERATION;
    }

    GL_get_color_src0(d3dp, dwStage, cop, &csrc0, &copn0);
    GL_get_color_src1(d3dp, dwStage, cop, &csrc1, &copn1);
    GL_get_color_src2(d3dp, dwStage, cop, &csrc2, &copn2);
    GL_get_alpha_src0(d3dp, dwStage, aop, &asrc0, &aopn0);
    GL_get_alpha_src1(d3dp, dwStage, aop, &asrc1, &aopn1);
    GL_get_alpha_src2(d3dp, dwStage, aop, &asrc2, &aopn2);

    if (D3D_GL_have_tex_ext(priv, EXT_texture_env_combine)) {
	if (cop->dwFlags & GLTF_NC) goto no_combiner;
	if ((cop->dwFlags & GLTF_C) || (aop->dwFlags & GLTF_C)) {
	    /* if we use the texture constant color ourselves, the app can't use it */
	    if (csrc0 == GL_CONSTANT_EXT ||
		csrc1 == GL_CONSTANT_EXT ||
		csrc2 == GL_CONSTANT_EXT) return D3DERR_UNSUPPORTEDCOLORARG;
	    if (asrc0 == GL_CONSTANT_EXT ||
		asrc1 == GL_CONSTANT_EXT ||
		asrc2 == GL_CONSTANT_EXT) return D3DERR_UNSUPPORTEDALPHAARG;
	}
	if ((cop->dwFlags & GLTF_SA) && (cop == aop)) {
	    /* if the alpha op is an extension of the color op, alpha args must match color args */
	    if (csrc0 != asrc0 || copn0 != aopn0 ||
		csrc1 != asrc1 || copn1 != aopn1 ||
		csrc2 != asrc2 || copn2 != aopn2) return D3DERR_UNSUPPORTEDALPHAARG;
	}
	return D3D_OK;
    }

no_combiner:
    /* we don't have texture combiners, validate what can be expressed without them */

    if (dwStage == 0) {
	if (csrc0 == GL_PRIMARY_COLOR_EXT) csrc0 = GL_PREVIOUS_EXT;
	if (csrc1 == GL_PRIMARY_COLOR_EXT) csrc1 = GL_PREVIOUS_EXT;
	if (csrc2 == GL_PRIMARY_COLOR_EXT) csrc2 = GL_PREVIOUS_EXT;
	if (asrc0 == GL_PRIMARY_COLOR_EXT) asrc0 = GL_PREVIOUS_EXT;
	if (asrc1 == GL_PRIMARY_COLOR_EXT) asrc1 = GL_PREVIOUS_EXT;
	if (asrc2 == GL_PRIMARY_COLOR_EXT) asrc2 = GL_PREVIOUS_EXT;
    }

    /* REPLACE */
    if (cop->cmode == GL_REPLACE) {
	if (csrc0 != GL_TEXTURE && csrc0 != GL_PREVIOUS_EXT) return D3DERR_UNSUPPORTEDCOLORARG;
	if (copn0 != GL_SRC_COLOR) return D3DERR_UNSUPPORTEDCOLORARG;
	if (aop->cmode != GL_REPLACE) return D3DERR_UNSUPPORTEDALPHAOPERATION;
	if (asrc0 != csrc0) return D3DERR_UNSUPPORTEDALPHAARG;
	if (aopn0 != GL_SRC_ALPHA) return D3DERR_UNSUPPORTEDALPHAARG;
	return D3D_OK;
    }

    /* MODULATE, BLEND, DECAL and ADD */
    if (cop->cmode == GL_BLEND || cop->cmode == GL_DECAL) {
	if (!(csrc0 == GL_TEXTURE && csrc1 == GL_PREVIOUS_EXT))
	    return D3DERR_UNSUPPORTEDCOLORARG;
    }
    else {
	if (!((csrc0 == GL_TEXTURE && csrc1 == GL_PREVIOUS_EXT) ||
	      (csrc1 == GL_TEXTURE && csrc0 == GL_PREVIOUS_EXT)))
	    return D3DERR_UNSUPPORTEDCOLORARG;
    }
    if (copn0 != GL_SRC_COLOR || copn1 != GL_SRC_COLOR)
	return D3DERR_UNSUPPORTEDCOLORARG;

    if (aop->cmode == GL_REPLACE) {
	/* FIXME: validate this only for textures without alpha channel
	 * (except when GL_DECAL is used) */
	if (asrc0 != GL_PREVIOUS_EXT) return D3DERR_UNSUPPORTEDALPHAARG;
	if (aopn0 != GL_SRC_ALPHA) return D3DERR_UNSUPPORTEDALPHAARG;
	return D3D_OK;
    }
    else if (aop->cmode == GL_MODULATE) {
	/* FIXME: validate this only for textures with alpha channel */
	if (!((asrc0 == GL_TEXTURE && asrc1 == GL_PREVIOUS_EXT) ||
	      (asrc1 == GL_TEXTURE && asrc0 == GL_PREVIOUS_EXT)))
	    return D3DERR_UNSUPPORTEDALPHAARG;
	if (aopn0 != GL_SRC_ALPHA || aopn1 != GL_SRC_ALPHA)
	    return D3DERR_UNSUPPORTEDALPHAARG;
	return D3D_OK;
    }
    else return D3DERR_UNSUPPORTEDALPHAOPERATION;
}

void GL_LoadTextureFactor(GL_D3D_priv* priv, GLfloat col[4])
{
    LPD3D_P d3dp = priv->d3dp;
    unsigned u;

    memcpy(priv->texture_factor, col, sizeof(GLfloat)*4);
    for (u=0; u<priv->lpCaps->wMaxTextureBlendStages; u++) {
	GL_TEXOP *cop = D3D_GL_get_tex_op(priv, d3dp->texstagestate[u][D3DTSS_COLOROP]);
	GL_TEXOP *aop = D3D_GL_get_tex_op(priv, d3dp->texstagestate[u][D3DTSS_ALPHAOP]);
	if (!cop) continue;
	if ((cop->dwFlags & GLTF_C) || (aop && (aop->dwFlags & GLTF_C)))
	    continue;
	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, col);
    }
}

HRESULT D3D_GL_texture_stage_state(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2TEXTURESTAGESTATE dat = (LPD3DHAL_DP2TEXTURESTAGESTATE)data1;
    LPD3D_P d3dp = priv->d3dp;
    if (dat->wStage >= priv->lpCaps->wMaxTextureBlendStages) {
	WARN("unsupported texture stage\n");
	return D3D_OK;
    }
    TRACE("stage %d, state %d, value %08lx\n", dat->wStage, dat->TSState, dat->dwValue);
    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + dat->wStage);
    switch (dat->TSState) {
    case D3DTSS_TEXTUREMAP:
#ifndef WIREFRAME
	if (dat->dwValue) {
#else
	if (dat->dwValue && !gWireframe) {
#endif   
            LPD3D_T tex = (LPD3D_T)dat->dwValue;
	    GLfloat col[4];
	    if (!TGLDEVICE(tex)) {
		HRESULT hr = GL_PrepTexture(priv, tex);
		if (FAILED(hr)) return hr;
	    }
	    TRACE("binding texture ID %d\n", TEXTURE_ID(tex));
	    glBindTexture(GL_TEXTURE_2D, TEXTURE_ID(tex));
	    if (GL_IsStageEnabled(priv, dat->wStage)) glEnable(GL_TEXTURE_2D);

	    /* since I think these parameters are bound to the texture in OpenGL,
	     * but not in Direct3D, they must be reloaded for every texture change */
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			    GL_mag[d3dp->texstagestate[dat->wStage][D3DTSS_MAGFILTER]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			    GL_mip_min[d3dp->texstagestate[dat->wStage][D3DTSS_MIPFILTER]]
				      [d3dp->texstagestate[dat->wStage][D3DTSS_MINFILTER]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
			    GL_tex_addr[d3dp->texstagestate[dat->wStage][D3DTSS_ADDRESSU]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
			    GL_tex_addr[d3dp->texstagestate[dat->wStage][D3DTSS_ADDRESSV]]);
	    GL_RS2Col(col, d3dp->texstagestate[dat->wStage][D3DTSS_BORDERCOLOR]);
	    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, col);
	    if (HAVE_GL_EXT(priv, EXT_texture_filter_anisotropic))
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
				d3dp->texstagestate[dat->wStage][D3DTSS_MAXANISOTROPY]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL,
			    d3dp->texstagestate[dat->wStage][D3DTSS_MAXMIPLEVEL]);
	} else {
	    glBindTexture(GL_TEXTURE_2D, 0);
	    glDisable(GL_TEXTURE_2D);
	}
	break;
    case D3DTSS_COLOROP:
    case D3DTSS_COLORARG1:
    case D3DTSS_COLORARG2:
    case D3DTSS_COLORARG0:
    case D3DTSS_ALPHAOP:
    case D3DTSS_ALPHAARG1:
    case D3DTSS_ALPHAARG2:
    case D3DTSS_ALPHAARG0:
	D3D_GL_set_tex_op(priv, dat->wStage, dat->TSState);
	break;
    case D3DTSS_TEXCOORDINDEX:
	priv->glFlags |= GLF_NEW_TEXCOORD_TRANSFORM;

	if ((d3dp->texstagestate[dat->wStage][D3DTSS_TEXCOORDINDEX]
	     & D3DTSS_TCI_MASK) != 0)
	{
	    priv->texture_flags &= ~GLTF_DEFAULT_TEXCOORD_TRANSFORM;
	}
	break;
    case 12: /* D3DTSS_ADDRESS */
    case D3DTSS_ADDRESSU:
    case D3DTSS_ADDRESSV:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
			GL_tex_addr[d3dp->texstagestate[dat->wStage][D3DTSS_ADDRESSU]]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
			GL_tex_addr[d3dp->texstagestate[dat->wStage][D3DTSS_ADDRESSV]]);
	break;
    case D3DTSS_BORDERCOLOR:
	{
	    GLfloat col[4];
	    GL_RS2Col(col, d3dp->texstagestate[dat->wStage][D3DTSS_BORDERCOLOR]);
	    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, col);
	}
	break;
    case D3DTSS_MAGFILTER:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_mag[d3dp->texstagestate[dat->wStage][D3DTSS_MAGFILTER]]);
	break;
    case D3DTSS_MINFILTER:
    case D3DTSS_MIPFILTER:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_mip_min[d3dp->texstagestate[dat->wStage][D3DTSS_MIPFILTER]]
				  [d3dp->texstagestate[dat->wStage][D3DTSS_MINFILTER]]);
	break;

    case D3DTSS_MIPMAPLODBIAS:
	/* FIXME: Reactivate this when DRI gets fixed - crasher in Mesa 3.4/3.5 gl_texture_pixels
	 * function that tries to dereference lambda[] array when lod_bias is on but mipmapping
	 * is off.
	 
	if (HAVE_GL_EXT(priv, EXT_texture_lod_bias))
	    glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT,
		      GL_RSF(dat->dwValue));
	*/ 
	break;
    case D3DTSS_MAXANISOTROPY:
	if (HAVE_GL_EXT(priv, EXT_texture_filter_anisotropic))
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
			    d3dp->texstagestate[dat->wStage][D3DTSS_MAXANISOTROPY]);
	break;
    case D3DTSS_MAXMIPLEVEL:
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, d3dp->texstagestate[dat->wStage][D3DTSS_MAXMIPLEVEL]);
        break; 
    case D3DTSS_TEXTURETRANSFORMFLAGS:
	priv->glFlags |= GLF_NEW_TEXCOORD_TRANSFORM;

	if (d3dp->texstagestate[dat->wStage][D3DTSS_TEXTURETRANSFORMFLAGS]
	    != D3DTTFF_DISABLE)
	    priv->texture_flags &= ~GLTF_DEFAULT_TEXCOORD_TRANSFORM;
        break; /* This is handled elsewhere - just record it for now */
    default:
	FIXME("unsupported texture stage state %d\n", dat->TSState);
	break;
    }
    CHECK_GL_ERROR("TextureStageState");
    return D3D_OK;
}

HRESULT WINAPI D3DGL_ValidateDevice(GL_D3D_priv* priv, LPDWORD lpdwPasses)
{
    LPD3D_P d3dp = priv->d3dp;
    unsigned u;
    HRESULT hr = D3D_OK;

    TRACE("(%p)->(%p)\n", priv, lpdwPasses);
    *lpdwPasses = 1;

    for (u=0; (!hr) && u < priv->lpCaps->wMaxTextureBlendStages &&
	      d3dp->texstagestate[u][D3DTSS_COLOROP] > D3DTOP_DISABLE; u++)
	hr = D3D_GL_validate_tex_stage(priv, u);

    TRACE("returning %ld\n", hr);
    return hr;
}
