/* Direct3D Vertex Buffer Implementation
   Copyright 2000-2001 TransGaming Technologies Inc.   
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include <d3d.h>

#include "debugtools.h"
#include "ddraw_private.h"
#include "d3d_private.h"
#include "d3dvbuffer/main.h"

static inline void Main_D3D_AddVertexBuffer(LPVOID pD3D, LPVOID This)
{
}

static inline void Main_D3D_RemoveVertexBuffer(LPVOID pD3D, LPVOID This)
{
}

DEFAULT_DEBUG_CHANNEL(ddraw);

#define THIS_VAR \
	ICOM_THIS_FROM(IDirect3DVertexBufferImpl, IDirect3DVertexBuffer7, \
		       iface)

static ICOM_VTABLE(IDirect3DVertexBuffer7) Main_IDirect3DVertexBuffer7_VTable;

static DWORD size_from_fvf(DWORD fvf)
{
    /* # of floats for position data */
    static const BYTE coord_sizes[] =
	{ 0 * sizeof(float), 3 * sizeof(float), 4 * sizeof(float),
	  4 * sizeof(float), 5 * sizeof(float), 6 * sizeof(float),
	  7 * sizeof(float), 8 * sizeof(float) };

    DWORD size;

    size = coord_sizes[(fvf & D3DFVF_POSITION_MASK) >> 1];

    if (fvf & D3DFVF_NORMAL) size += 3 * sizeof(float);
    if (fvf & D3DFVF_DIFFUSE) size += sizeof(DWORD);
    if (fvf & D3DFVF_SPECULAR) size += sizeof(DWORD);

    if (fvf & 0xFFFF0000)
    {
	/* use slow and probably wrong algorithm */
	DWORD i;

	DWORD texcount = (fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT;

	for (i = 0; i < texcount && i < 4; i++)
	{
	    static const BYTE texcoord_sizes[] =
		{ 2 * sizeof(float), 3 * sizeof(float), 1 * sizeof(float),
		  4 * sizeof(float) };

	    size += texcoord_sizes[(fvf >> (2*i + 16)) & 3];
	}

	size += 2 * sizeof(float) * (texcount - 4);
    }
    else
    {
	size += ((fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)
	    * 2 * sizeof(float);
    }

    return size;
}

HRESULT Main_D3DVertexBuffer_Construct(IDirect3DVertexBufferImpl* This,
				       IDirect3D7* pD3D,
				       LPD3DVERTEXBUFFERDESC desc,
				       DWORD dwFlags)
{
    TRACE("%p %lu, 0x%lx, 0x%lx, %lu\n", This, desc->dwSize, desc->dwCaps,
	  desc->dwFVF, desc->dwNumVertices);

    if (dwFlags != 0)
	return DDERR_INVALIDPARAMS;

    ICOM_INIT_INTERFACE(This, IDirect3DVertexBuffer7,
			Main_IDirect3DVertexBuffer7_VTable);

    This->desc    = *desc;
    This->flags   = dwFlags;
    This->owner   = (IDirect3D7Impl*)pD3D; /* XXX */
    This->vertex_size = size_from_fvf(This->desc.dwFVF);
    This->vb_size = This->vertex_size * This->desc.dwNumVertices;

    This->final_release = Main_D3DVertexBuffer_final_release;
    This->draw = Main_D3DVertexBuffer_draw;
    This->draw_indexed = Main_D3DVertexBuffer_draw_indexed;

    This->ref = 1;

    This->vb_data = VirtualAlloc(NULL, This->vb_size, MEM_RESERVE | MEM_COMMIT,
				 PAGE_READWRITE);

    Main_D3D_AddVertexBuffer(pD3D, This);

    return D3D_OK;
}

HRESULT Main_D3DVertexBuffer_Create(IDirect3D7* pD3D,
				    LPD3DVERTEXBUFFERDESC desc, DWORD dwFlags,
				    IDirect3DVertexBuffer7** ppVB)
{
    IDirect3DVertexBufferImpl* This;
    HRESULT hr;

    This = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(*This));
    if (This == NULL) return E_OUTOFMEMORY;

    This->private = NULL;

    hr = Main_D3DVertexBuffer_Construct(This, pD3D, desc, dwFlags);
    if (FAILED(hr))
	HeapFree(GetProcessHeap(), 0, This);
    else
	*ppVB = ICOM_INTERFACE(This, IDirect3DVertexBuffer7);

    return hr;
}

void Main_D3DVertexBuffer_final_release(IDirect3DVertexBufferImpl* This)
{
    Main_D3D_RemoveVertexBuffer(This->owner, This);
    VirtualFree(This->vb_data, This->vb_size, MEM_RELEASE);
}

void Main_D3DVertexBuffer_Destroy(IDirect3DVertexBufferImpl* This)
{
    This->final_release(This);
    if (This->private != NULL && This->private != This+1)
	HeapFree(GetProcessHeap(), 0, This->private);
    HeapFree(GetProcessHeap(), 0, This);
}

void Main_D3DVertexBuffer_ForceDestroy(IDirect3DVertexBufferImpl* This)
{
    WARN("destroying VB %p with refcnt %lu\n", This, This->ref);
    Main_D3DVertexBuffer_Destroy(This);
}

ULONG WINAPI Main_D3DVertexBuffer_AddRef(IDirect3DVertexBuffer7* iface)
{
    THIS_VAR;

    return ++This->ref;
}

ULONG WINAPI Main_D3DVertexBuffer_Release(IDirect3DVertexBuffer7* iface)
{
    THIS_VAR;

    if (--This->ref == 0)
    {
	Main_D3DVertexBuffer_Destroy(This);
	return 0;
    }

    return This->ref;
}

HRESULT WINAPI
Main_D3DVertexBuffer_QueryInterface(IDirect3DVertexBuffer7* iface, REFIID riid,
				    LPVOID* ppObj)
{
    THIS_VAR;

    if (IsEqualGUID(riid, &IID_IDirect3DVertexBuffer7)
	|| IsEqualGUID(riid, &IID_IUnknown))
    {
	++This->ref;
	*ppObj = ICOM_INTERFACE(This, IDirect3DVertexBuffer7);
	return S_OK;
    }
    else
    {
	/* e.g. IDirect3DVertexBuffer and maybe IDirectDrawSurface* */
	FIXME("(%s) unsupported interface\n", debugstr_guid(riid));
	return E_NOINTERFACE;
    }
}
				    

HRESULT WINAPI
Main_D3DVertexBuffer_GetVertexBufferDesc(IDirect3DVertexBuffer7* iface,
					 LPD3DVERTEXBUFFERDESC pVBDesc)
{
    THIS_VAR;

    memcpy(pVBDesc, &This->desc, pVBDesc->dwSize);

    return D3D_OK;
}

HRESULT WINAPI
Main_D3DVertexBuffer_Lock(IDirect3DVertexBuffer7* iface, DWORD dwFlags,
			  LPVOID* ppData, LPDWORD pdwSize)
{
    THIS_VAR;
    TRACE("%p 0x%lx %p %p\n", iface, dwFlags, ppData, pdwSize);

    *ppData = This->vb_data;

    if (pdwSize != NULL)
	*pdwSize = This->vb_size;

    This->locked = TRUE;

    return D3D_OK;
}

HRESULT WINAPI
Main_D3DVertexBuffer_Unlock(IDirect3DVertexBuffer7* iface)
{
    THIS_VAR;
    This->locked = FALSE;
    return D3D_OK;
}

HRESULT WINAPI
Main_D3DVertexBuffer_ProcessVertices(IDirect3DVertexBuffer7* iface,
				     DWORD dwVertexOp, DWORD dwDestIndex,
				     DWORD dwCount,
				     LPDIRECT3DVERTEXBUFFER7 pSrc,
				     DWORD dwSrcIndex,
				     LPDIRECT3DDEVICE7 pD3DDevice,
				     DWORD dwFlags)
{
    THIS_VAR;
    IDirect3DVertexBufferImpl  *pSrcVbuffer = (IDirect3DVertexBufferImpl *)pSrc;
    TRACE("%p -> %p Vop: %lu, DestIdx: %lu, Count: %lu, SrcIdx: %lu Flags: %lu\n", 
	  pSrc, iface, dwVertexOp, dwDestIndex, dwCount, dwSrcIndex, dwFlags);

    /* For now, we don't do anything at all with the data, we just copy 
     * it all into the dst.  We hope that the app doesn't try to read 
     * from the data, and when DrawPrimitiveVB gets called, we just 
     * assume untransformed unlit unclipped vertices, regardless of 
     * what the vertex buffer format is.   */

    /* Force dst fvf to the same type as src */
    This->vertex_size = size_from_fvf(pSrcVbuffer->desc.dwFVF);
    This->desc.dwFVF = pSrcVbuffer->desc.dwFVF;    
    
    if (This->vb_size < (This->vertex_size * This->desc.dwNumVertices))
    {
        VirtualFree(This->vb_data, This->vb_size, MEM_RELEASE);
        This->vb_size = This->vertex_size * This->desc.dwNumVertices;

        This->vb_data = VirtualAlloc(NULL, This->vb_size, MEM_RESERVE | MEM_COMMIT,
				 PAGE_READWRITE);
    }
   
    memcpy((char*)(This->vb_data) + (This->vertex_size * dwDestIndex),
    	   (char*)(pSrcVbuffer->vb_data) + (pSrcVbuffer->vertex_size * dwSrcIndex),
    	   (This->vertex_size * dwCount));

    return D3D_OK;
/*    
    ERR("\n");
    abort();
*/
}

HRESULT WINAPI
Main_D3DVertexBuffer_ProcessVerticesStrided(IDirect3DVertexBuffer7* iface,
					    DWORD dwVertexOp,
					    DWORD dwDestIndex, DWORD dwCount,
					    LPD3DDRAWPRIMITIVESTRIDEDDATA pStrideData,
					    DWORD dwFVF,
					    LPDIRECT3DDEVICE7 pD3DDevice,
					    DWORD dwFlags)
{
    ERR("\n");
    abort();
}

HRESULT WINAPI
Main_D3DVertexBuffer_Optimize(IDirect3DVertexBuffer7* iface,
			      LPDIRECT3DDEVICE7 pD3DDevice, DWORD dwFlags)
{
    THIS_VAR;

    if (This->optimized)
	return D3DERR_VERTEXBUFFEROPTIMIZED;

    if (This->locked)
	return D3DERR_VERTEXBUFFERLOCKED;

    /* do nothing */

    return D3D_OK;
}

static inline LPVOID vertex_index(IDirect3DVertexBufferImpl* This,
				  DWORD index)
{
    return (char *)This->vb_data + index * This->vertex_size;
}

HRESULT
Main_D3DVertexBuffer_draw(IDirect3DVertexBufferImpl* This,
			  IDirect3DDevice7* pD3D,
			  D3DPRIMITIVETYPE d3dptPrimitiveType,
			  DWORD dwStart, DWORD dwCount, DWORD dwFlags)
{
    TRACE("\n");
    return IDirect3DDevice7_DrawPrimitive(pD3D, d3dptPrimitiveType,
					  This->desc.dwFVF,
					  vertex_index(This, dwStart), dwCount,
					  dwFlags);
}

HRESULT
Main_D3DVertexBuffer_draw_indexed(IDirect3DVertexBufferImpl* This,
				  IDirect3DDevice7* pD3D,
				  D3DPRIMITIVETYPE d3dptPrimitiveType,
				  DWORD dwStart, DWORD dwVertexCount,
				  LPWORD pwIndices, DWORD dwIndexCount,
				  DWORD dwFlags)
{
    TRACE("\n");
    return IDirect3DDevice7_DrawIndexedPrimitive(pD3D, d3dptPrimitiveType,
						 This->desc.dwFVF,
						 This->vb_data,
						 This->desc.dwNumVertices,
						 pwIndices, dwIndexCount,
						 dwFlags);
}

static ICOM_VTABLE(IDirect3DVertexBuffer7) Main_IDirect3DVertexBuffer7_VTable =
{
    Main_D3DVertexBuffer_QueryInterface,
    Main_D3DVertexBuffer_AddRef,
    Main_D3DVertexBuffer_Release,
    Main_D3DVertexBuffer_Lock,
    Main_D3DVertexBuffer_Unlock,
    Main_D3DVertexBuffer_ProcessVertices,
    Main_D3DVertexBuffer_GetVertexBufferDesc,
    Main_D3DVertexBuffer_Optimize,
    Main_D3DVertexBuffer_ProcessVerticesStrided
};
