/* Wrapper for OpenGL includes...
     Copyright 1998 - Lionel Ulmer
     Copyright 2001 TransGaming Technologies, Inc.

   This wrapper is needed because Mesa uses also the CALLBACK / WINAPI
   constants.  It's also needed to work around certain common problems
   in Linux GL headers where key function pointers are not defined. */


#ifndef __WINE_WINE_GL_H
#define __WINE_WINE_GL_H

#ifndef __WINE_CONFIG_H 
# error You must include config.h to use this header 
#endif 

#if defined(HAVE_OPENGL)

#include "ts_xlib.h"

/* As GLX relies on X, this is needed */
#define ENTER_GL() wine_tsx11_lock()
#define LEAVE_GL() wine_tsx11_unlock()

#undef APIENTRY
#undef CALLBACK
#undef WINAPI

#define XMD_H /* This is to prevent the Xmd.h inclusion bug to happen :-/ */
#include <GL/gl.h>
#include <GL/glu.h>
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>

/* Work around a gl.h vs glext.h snafu - gl.h may activate defines
   that prevent glext.h from giving us function pointer prototypes
   for the extensions below (which we need). If these defines are
   active, we have to define these prototypes ourselves. Of course
   this is further complicated by the fact that Mesa's gl.h header
   (if it's around instead of SGI's gl.h) includes glext.h itself.
   Hence the checks for GLEXT_VERSION for everything but 
   GL_ARB_multitexture (Mesa gl.h defines GL_ARB_multitexture
   itself, rather than leaving it up to glext.h - aak!) */  

#if defined(GL_ARB_multitexture) 
#define WINE_DDRAW_need_multitexture_pfns
#endif 

#if defined(GL_EXT_secondary_color) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_secondary_color_pfns
#endif 

#if defined(GL_ARB_texture_compression) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_texture_compression_pfns
#endif

#if defined(GL_NV_vertex_array_range) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_NV_VAR_pfns
#endif

#if defined(GL_NV_fence) && !defined(GL_GLEXT_VERSION)
#define WINE_DDRAW_need_NV_fence_pfns
#endif

#ifdef HAVE_GL_GLEXT_H
#include <GL/glext.h>
#endif

#ifndef APIENTRY
#define APIENTRY
#endif

#ifdef WINE_DDRAW_need_multitexture_pfns
typedef void (APIENTRY *PFNGLACTIVETEXTUREARBPROC)(GLenum texture);
typedef void (APIENTRY *PFNGLCLIENTACTIVETEXTUREARBPROC)(GLenum texture);
typedef void (APIENTRY *PFNGLMULTITEXCOORD2FARBPROC)(GLenum target, GLfloat u, GLfloat v);
#endif
 
#ifdef WINE_DDRAW_need_secondary_color_pfns
typedef void (APIENTRY * PFNGLSECONDARYCOLOR3UBEXTPROC) (GLubyte red, GLubyte green, GLubyte blue);
typedef void (APIENTRY * PFNGLSECONDARYCOLORPOINTEREXTPROC) (GLint size, GLenum type,
              GLsizei stride, GLvoid *pointer);
#endif
 
#ifdef WINE_DDRAW_need_texture_compression_pfns
typedef void (APIENTRY * PFNGLCOMPRESSEDTEXIMAGE2DARBPROC) (GLenum target, GLint level,
              GLenum internalformat, GLsizei width, GLsizei height, GLint border,
              GLsizei imageSize, const GLvoid *data);
typedef void (APIENTRY * PFNGLCOMPRESSEDTEXSUBIMAGE2DARBPROC) (GLenum target, GLint level,
              GLint xoffset, GLint yoffset, GLsizei width, GLsizei height,
              GLenum format, GLsizei imageSize, const GLvoid *data);
#endif


#ifdef WINE_DDRAW_need_NV_VAR_pfns
typedef void (APIENTRY * PFNGLFLUSHVERTEXARRAYRANGENVPROC) (void);
typedef void (APIENTRY * PFNGLVERTEXARRAYRANGENVPROC) (GLsizei size, const GLvoid *pointer);
#endif

#undef  XMD_H

#undef APIENTRY
#undef CALLBACK
#undef WINAPI

/* Redefines the constants */
#define CALLBACK    __stdcall
#define WINAPI      __stdcall
#define APIENTRY    WINAPI

#endif /* HAVE_OPENGL */

#endif /* __WINE_WINE_GL_H */
